/*
 * Copyright (c) 2016, Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Pushkar Singh <psingh@sangoma.com>
 *
 */

/********************************************************************************/
/*                                                                              */
/*                           libraries inclusion                                */
/*                                                                              */
/********************************************************************************/
#include "sample_tapping.h"

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
#ifndef __WINDOWS__
void on_sigint(int signal);
void on_sigusr(int signal);
#endif

int exit_sample_app(int max_wait);

int open_audio_file(uint32_t chan_id, uint32_t span_id, void *obj);

/* Application LOGGING MODULE FUNCTION Declaration*/
ftdm_status_t app_open_log_file(void);

/* Set SDK_DEBUGGING since this is a sample application,
 * This should be turned off in Production */
#define SDK_DEBUGGING 1
#define MAX_FILE_PATH_SIZE 25

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;
int g_init = 0;
FILE *log_file_ptr;

span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE + 1] = { {0} };

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

/* Logger Helper Function *******************************************************/

#ifdef __WINDOWS__
#include <ftdm_declare.h>
typedef struct ftdm_timezone {
	int tz_minuteswest;                     /* minutes W of Greenwich */
	int tz_dsttime;                         /* type of dst correction */
};

/**********************************************************************************
 * Fun   : gettimeofday()
 *
 * Desc  : To get the time of the Day in hh:mm:ss format 
 *
 * Ret   : int
**********************************************************************************/
static inline int gettimeofday(struct timeval *tv, struct ftdm_timezone *tz)
{
	FILETIME ft;
	unsigned __int64 tmpres = 0;
	static int tzflag;
	if (NULL != tv) {
		GetSystemTimeAsFileTime(&ft);
		tmpres |= ft.dwHighDateTime;
		tmpres <<= 32;
		tmpres |= ft.dwLowDateTime;

		/*converting file time to unix epoch */
		tmpres /= 10;                   /*convert into microseconds */
		tmpres -= DELTA_EPOCH_IN_MICROSECS;
		tv->tv_sec = (long)(tmpres / 1000000UL);
		tv->tv_usec = (long)(tmpres % 1000000UL);
	}
	if (NULL != tz) {
		if (!tzflag) {
			_tzset();
			tzflag++;
		}
		tz->tz_minuteswest = _timezone / 60;
		tz->tz_dsttime = _daylight;
	}
	return 0;
}
#endif

/**********************************************************************************
 * Fun   : sangoma_get_timestamp_str()
 *
 * Desc  : To get the exact timerstamp with date month and year too 
 *
 * Ret   : void
**********************************************************************************/
static inline void sangoma_get_timestamp_str(char *timestamp)
{
#ifdef __WINDOWS__
	__time64_t tv; 
#else
	struct timeval tv; 
#endif
	struct tm *ptm;
	char time_string[128];
	long milliseconds;

	if (!timestamp) {
		return;
	}   

	/* Obtain the time of day, and convert it to a tm struct. */
#ifdef __WINDOWS__
	_time64 (&tv);
	ptm = _localtime64(&tv);
#else
	gettimeofday(&tv, NULL);
	ptm = localtime(&tv.tv_sec);
#endif
	/* Format the date and time, down to a single second. */
	strftime(time_string, sizeof(time_string), "%Y-%m-%d-%H-%M-%S", ptm);

	/* Compute milliseconds from microseconds. */
#ifdef __WINDOWS__
	milliseconds = 0;
#else
	milliseconds = tv.tv_usec / 1000;
#endif
	/* Print the formatted time, in seconds, followed by a decimal point and the milliseconds. */
	snprintf(timestamp, 80, "%s.%03ld", time_string, milliseconds);

	return;
}
/*********************************************************************************/

/* XML Parsing Helper Function ***************************************************/
/**********************************************************************************
 * Fun   : xml_child()
 *
 * Desc  : To get the child node from the main xml node passed i.e. from the whole
 * 	   xml configuration file get the specific set of configuration as passed
 *
 * Ret   : void
**********************************************************************************/
static xmlNode *xml_child(xmlNode *node, const char *child_name)
{
	xmlNode *cur_node = NULL;
	for (cur_node = node->children; cur_node; cur_node = cur_node->next) {
		fprintf(stdout, "node %s = %s\n", cur_node->name, child_name);
		if (cur_node->type == XML_ELEMENT_NODE && !xmlStrcmp(cur_node->name, (xmlChar *)child_name)) {
			fprintf(stdout, "node %s found in node %p\n", child_name, node);
			return cur_node;
		}   
	}   
	fprintf(stdout, "node %s not found in node %p\n", child_name, node);
	return NULL;
}

/**********************************************************************************
 * Fun   : add_isdn_tap_profile_parameters()
 *
 * Desc  : To get the confiuration parameters and their values as present in xml
 * 	   xml file and store the information in order to pass it to the respective
 * 	   module
 *
 * Ret   : (int)The index of parameter that are stored in array till function ends
**********************************************************************************/
static int add_isdn_tap_profile_parameters(xmlNode *cfg, const char *profname, ftdm_conf_parameter_t *parameters, int len)
{
	xmlNode *profnode = NULL, *profile = NULL, *param = NULL;
	int paramindex = 0;

	profnode = xml_child(cfg, "config_profiles");

	if (!profnode) {
		fprintf(stderr, "cannot find profile '%s', there is no 'config_profiles' XML section\n", profname);
		return 0;
	}

	/* search the profile */
	for (profile = xml_child(profnode, "profile"); profile; profile = profile->next) {
		char *name = (char *) xmlGetProp(profile, (xmlChar *)"name");
		if (!name) {
			continue;
		}
		if (!strcasecmp(name, profname)) {
			break;
		}
	}

	if (!profile) {
		fprintf(stderr, "failed to find profile '%s'\n", profname);
		return 0;
	}

	for (param = xml_child(profile, "param"); param; param = param->next) {
		char *var = (char *) xmlGetProp(param, (xmlChar *)"name");
		char *val = (char *) xmlGetProp(param, (xmlChar *)"value");
		if (!var || !val) {
			continue;
		}
		parameters[paramindex].var = var;
		parameters[paramindex].val = val;
		paramindex++;
	}

	return paramindex;
}
/*********************************************************************************/

/* SIGNAL handling Function ******************************************************/
/**********************************************************************************
 * Fun   : on_sigusr()
 *
 * Desc  : In case if there is any iterruption by user in order then stop the 
 * 	   application and exit
 *
 * Ret   : void
**********************************************************************************/
#ifndef __WINDOWS__
void on_sigusr(int signal)
{
	switch (signal) {
	case SIGUSR1:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		break;

	case SIGUSR2:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		break;
	}

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA Tapping Module. Application is getting closed\n");
	getchar();
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");

	exit(EXIT_FAILURE);
}

/**********************************************************************************
 * Fun   : on_sigint()
 *
 * Desc  : In case if SIGINT is received then stop the application and exit
 *
 * Ret   : void
**********************************************************************************/
void on_sigint(int signal)
{
	g_running = 0;

	ftdm_log(FTDM_LOG_INFO, "Signal[%d] recieved ...\n", signal);

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA Tapping Module. Application is getting closed\n");
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");
	getchar();

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");

	exit(EXIT_FAILURE);
}

/**********************************************************************************
 * Fun   : signal_handler()
 *
 * Desc  : Signal Handler function and execution proper signal handling technique
 *	   depending up which signal is received.
 *
 * Ret   : void
**********************************************************************************/
static void signal_handler (int sig, siginfo_t *si, void *uc)
{
	printf("si_signo[%d], si_errno[%d] , si_code[%d],  si_pid[%d], si_uid[%d], si_status[%d], si_addr[%d], si_fd[%d] \n",
			si->si_signo, si->si_errno, si->si_code,
			si->si_pid, si->si_uid, si->si_status,
			(int)si->si_addr, si->si_fd);

	switch (sig)
	{
		case SIGTERM:
			{
				printf("signal_handler: SIGTERM received \n");
				on_sigint(SIGTERM);
				break;
			}
		case SIGKILL:
			{
				printf("signal_handler: SIGKILL received \n");
				on_sigint(SIGKILL);
				break;
			}

		case SIGUSR1:
			{
				printf("signal_handler: SIGUSR1 received \n");
				on_sigusr(SIGUSR1);
				break;
			}
		case SIGUSR2:
			{
				printf("signal_handler: SIGUSR2 received \n");
				on_sigusr(SIGUSR2);
				break;
			}
		case SIGINT:
			{
				printf("signal_handler: SIGINT received \n");
				on_sigint(SIGINT);
				break;
			}

		default:
			printf("signal_handler: Unknown[%d] signal received \n", sig);
			break;
	}
}

/**********************************************************************************
 * Fun   : add_signal_handlers()
 *
 * Desc  : Add proper signal handler in order handle various signal properly 
 *
 * Ret   : void
**********************************************************************************/
void add_signal_handlers(int signal)
{
	struct sigaction sa;

	memset(&sa,0,sizeof(sa));
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = signal_handler;
	sigemptyset(&sa.sa_mask);
	sigaction(signal, &sa, NULL);
}
#endif

/*********************************************************************************/

/* FTDM Command Helper Functions *************************************************/
/**********************************************************************************
 * Fun   : sample_app_exec_io_stats()
 *
 * Desc  : Execute IO status command
 *
 * Ret   : void
**********************************************************************************/
static void sample_app_exec_io_stats(const char* cmd)
{
	int i=0;
	if (!cmd) 
		return;
	for (i = 1;i<FTDM_MAX_SPANS_INTERFACE;i++) {
		if (!SPANS[i].ftdm_span) continue;
		sangoma_freetdm_exec_io_stats(cmd, SPANS[i].ftdm_span);
	}
}

/**********************************************************************************
 * Fun   : sample_app_set_rx_queuesize()
 *
 * Desc  : Set span signalling channel RX queue size
 *
 * Ret   : void
**********************************************************************************/
void sample_app_set_rx_queuesize(int rxqueue_size)
{
	int i=0; 
	for (i = 1;i<FTDM_MAX_SPANS_INTERFACE;i++) {
		if (!SPANS[i].ftdm_span) continue;
		sangoma_freetdm_set_span_sig_queuesize(SPANS[i].ftdm_span, rxqueue_size); 
	}    
}
/*********************************************************************************/

/* Tapping Aplication Functions **************************************************/
/**********************************************************************************
 * Fun   : tapping_main_menu()
 *
 * Desc  : Provides option in order to stop the application at any point of time
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
int tapping_main_menu(void)
{
	char choice[6];

	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "	    SANGOMA PRI/BRI TAPPING MODULE STARTED             \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Please enter your choice:  \n");
	fprintf(stdout, "1: Calls Statistics \n");
	fprintf(stdout, "2: Exit \n");
	fprintf(stdout, "Please enter 1 at any point of time if you want to exit from application \n");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			return FTDM_FAIL;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Exit from application
		 */
		switch (choice[0]) {
		case '1':
			sample_app_exec_io_stats("stats");
			sangoma_freetdm_isdn_tap_statistics();
			
			break;

		case '2':
			fprintf(stdout, "SANGOMA Tapping Sample Application is getting closed...\n");

			/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
			 * the configired wanpipes/hwports */
			exit_sample_app(20);

			fprintf(stdout, "\n\n\n=========================================================\n");
			fprintf(stdout, "                      GOOD BYE                            \n");
			fprintf(stdout, "=========================================================\n");

			return FTDM_SUCCESS;

		default:
			break;
		}
	}

	return FTDM_FAIL;
}

/**********************************************************************************
 * Fun   : start_tapping()
 *
 * Desc  : Starts tapping the call on the respective span and channel on which 
 * 	   call is established
 *
 * Ret   : NULL
**********************************************************************************/
static void *start_tapping(ftdm_thread_t *me, void *obj)
{
	uint8_t frame[1024];
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	ftdm_channel_t *chan = NULL;
	ftdm_span_t *span = NULL;
	ftdm_size_t len;
	uint32_t span_id = sigmsg->span_id;
	uint32_t chan_id = sigmsg->chan_id;
	ftdm_caller_data_t *caller_data = NULL;
	char timestamp[80];
	char path[512];
	ftdm_size_t dlen=0x00;
	int fd = -1;

	memset(path, 0 , sizeof(path));
	memset(timestamp, 0 , sizeof(timestamp));

	ftdm_log(FTDM_LOG_DEBUG, "Enter: start_tapping on span_id[%d] chanid[%d] \n", span_id, chan_id);

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid signalling message\n");
		goto done;
	}

	if (!span_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Span ID\n");
		goto done;
	}

	ftdm_span_find(span_id, &span);

	/* Send request to Freetdm to start tracing the media */
	if (!chan_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel ID\n");
		goto done;
	}

	if (!span) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid span found\n");
		goto done;
	}
	chan = ftdm_span_get_channel(span, chan_id);

	if (!chan) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel found\n");
		goto done;
	}

	ftdm_log(FTDM_LOG_DEBUG, "Start tapping on span %d and channel %d\n", span_id, chan_id);

	sangoma_get_timestamp_str(timestamp);

	caller_data = ftdm_channel_get_caller_data(chan);

	if (caller_data) {
		if (!ftdm_strlen_zero(caller_data->dnis.digits)) {
			sprintf(path, "recordings/%s-%d-%d-%s", timestamp, span_id, chan_id, caller_data->dnis.digits);
		} else {
			sprintf(path, "recordings/%s-%d-%d", timestamp, span_id, chan_id);
		}
	} else {
		sprintf(path, "recordings/%s-%d-%d", timestamp, span_id, chan_id);
	}

	if (!strlen(path)) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Path\n");
		goto done;
	}

	if ((fd = open_audio_file(chan_id, span_id, path)) == -1) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to start audio tracing on span[%d]-chan[%d]\n", span_id, chan_id);
		goto done;
	}

	/* Start reading the media, and write to a file in order to tap the call */ 
	while (ftdm_channel_call_check_answered(chan)) {
		ftdm_wait_flag_t flags = FTDM_READ;

		if (ftdm_channel_wait(chan, &flags, -1) == FTDM_FAIL) {
			break;
		}
		len = sizeof(frame);

		if (flags & FTDM_READ) {
			if (ftdm_channel_read(chan, frame, &len) == FTDM_SUCCESS) {
				/* Passive Tapping application should not write to channel, only read and write the data
				 * to a file */
				dlen = len;
				if ((ftdm_size_t)write(fd, frame, (int)dlen) != dlen) {
					ftdm_log(FTDM_LOG_WARNING, "Raw input trace failed to write all of the %"FTDM_SIZE_FMT" bytes\n", dlen);
				}
			}
		}

		/* No data to read..continue with checking..as if call terminate this loop will automatically end due to while condition */
	}

	if (fd >= 0) {
		ftdm_log(FTDM_LOG_DEBUG, "Stop Tapping and closing [%s] file with fd[%d], as call is hanged upon span[%d] chan[%d]\n", path, fd, span_id, chan_id);
		close(fd);
		fd = -1;
	}

  done:
	ftdm_safe_free(sigmsg->raw.data);
	ftdm_safe_free(sigmsg);
	return NULL;
}

/**********************************************************************************
 * Fun   : tap_handle_raw_message()
 *
 * Desc  : Starts tapping the call on the respective span and channel on which
 * 	   call is established
 *
 * Ret   : NULL
**********************************************************************************/
static void *tap_handle_raw_message(ftdm_thread_t *me, void *obj)
{
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	isdn_tap_msg_info_t *tap_msg = NULL;		/* tap raw message */
	uint32_t span_id = 0;				/* span id on which call is present */
	uint32_t chan_id = 0;				/* chan id on which call is present */
	uint32_t call_ref = 0;				/* call rference value of the call */
	tap_call_direction_t call_dir;			/* call direction i.e. incoming or outgoing */
	uint8_t call_con = 0;				/* to detect whether call is connected or not */
	char clg_num[25];				/* Calling party number if present */
	char cld_num[25];				/* Called party number if present */

	memset(cld_num, 0, sizeof(cld_num));
	memset(clg_num, 0, sizeof(clg_num));

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid signalling message\n");
		return NULL;
	}

	if (!(sigmsg->raw.data)) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid raw data present in signalling message\n");
		goto done;
	}

	tap_msg = sigmsg->raw.data;

	span_id = tap_msg->span_id;
	chan_id = tap_msg->chan_id;
	call_ref = tap_msg->call_ref;
	call_dir = tap_msg->call_dir;
	call_con = tap_msg->call_con;

	ftdm_log(FTDM_LOG_INFO, "*****************************************\n");
	ftdm_log(FTDM_LOG_INFO, "              Call Details               \n");
	ftdm_log(FTDM_LOG_INFO, "*****************************************\n");

	if (span_id) {
		ftdm_log(FTDM_LOG_INFO, "Call Direction : %s\n", ISDN_CALL_DIRECTION(call_dir));
	}

	ftdm_log(FTDM_LOG_INFO, "Call Received of Span ID : %d\n", span_id);
	ftdm_log(FTDM_LOG_INFO, "Call Received of Chan ID : %d\n", chan_id);
	ftdm_log(FTDM_LOG_INFO, "Call reference value : %d\n", call_ref);

	if (call_con) {
		ftdm_log(FTDM_LOG_INFO, "Call Type : Successfull Call\n");
	} else {
		ftdm_log(FTDM_LOG_INFO, "Call Type : Missed Call\n");
	}

	if (!ftdm_strlen_zero(tap_msg->calling_num)) {
		ftdm_set_string(clg_num, tap_msg->calling_num);
		ftdm_log(FTDM_LOG_INFO, "Calling Number : %s\n", clg_num);
	}

	if (!ftdm_strlen_zero(tap_msg->called_num)) {
		ftdm_set_string(cld_num, tap_msg->called_num);
		ftdm_log(FTDM_LOG_INFO, "Called Number : %s\n", cld_num);
	}

	ftdm_log(FTDM_LOG_INFO, "*****************************************\n\n");

	ftdm_safe_free(sigmsg->raw.data);
  done:
	ftdm_safe_free(sigmsg);
	return NULL;
}

/**********************************************************************************
 * Fun   : app_freetdm_event_queue_handler()
 *
 * Desc  : Starts respective io operation w.r.t different event received 
 *
 * Ret   : NULL
**********************************************************************************/
FIO_SIGNAL_CB_FUNCTION (on_isdn_tap_signal)
{
	unsigned int span_id, chan_id;
	ftdm_channel_t *ftdmchan;

	ftdmchan = sigmsg->channel;

	span_id = ftdm_channel_get_ph_span_id(ftdmchan);
	chan_id = ftdm_channel_get_ph_id(ftdmchan);

	ftdm_sigmsg_t *sigmsg_cpy = NULL;

	switch (sigmsg->event_id) {
		case FTDM_SIGEVENT_START:
			{
				sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
				memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
				ftdm_log(FTDM_LOG_DEBUG, "SIG_START event received. Starts io data operation for [s%dc%d]. \n",
						sigmsg->span_id, sigmsg->chan_id);

				/* Incoming Call. Proceed to answer the call and then
				   launch a media thread that will perform I/O on the channel */
				ftdm_channel_call_answer(sigmsg->channel);

				if (ftdm_thread_create_detached(start_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
					ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
							sigmsg->span_id, sigmsg->chan_id);
					break;
				}

				break;
			}

		case FTDM_SIGEVENT_UP:
			{
				sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
				memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
				ftdm_log(FTDM_LOG_DEBUG, "SIG_UP event received. Starts io data operation for [s%dc%d]. \n",
						sigmsg->span_id, sigmsg->chan_id);

				/* Outbound call has been answered, proceed to launch a meda thread that will
				   perform I/O on the channel */
				if (ftdm_thread_create_detached(start_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
					ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
							sigmsg->span_id, sigmsg->chan_id);
					break;
				}
				break;
			}

		case FTDM_SIGEVENT_RAW:
			{
				sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
				memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));

				if (sigmsg->raw.data) {
					sigmsg_cpy->raw.data = ftdm_malloc(sigmsg->raw.len);
					memcpy(sigmsg_cpy->raw.data, sigmsg->raw.data, sigmsg->raw.len);
				}

				ftdm_log(FTDM_LOG_DEBUG, "SIG_RAW event received\n");

				/* Handle Raw message received from lower layer */
				if (ftdm_thread_create_detached(tap_handle_raw_message, sigmsg_cpy) != FTDM_SUCCESS) {
					ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
							sigmsg->span_id, sigmsg->chan_id);
					break;
				}
				break;
			}

		default:
			break;
	}

	return FTDM_SUCCESS;
}

static int load_config(const char *cf)
{
	xmlDoc *doc = NULL;
	xmlNode *cfg, *param, *spans, *myspan;

	doc = xmlReadFile(cf, NULL, 0);
	if (!doc) {
		fprintf(stderr, "Failed to parse XML file '%s'\n", cf);
		return -1;

	}

	cfg = xmlDocGetRootElement(doc);

	if ((spans = xml_child(cfg, "sangoma_isdn_tap_spans"))) {
		for (myspan = xml_child(spans, "span"); myspan; myspan = myspan->next) {
			ftdm_status_t zstatus = FTDM_FAIL;
			ftdm_conf_parameter_t spanparameters[30];
			char *id = (char *) xmlGetProp(myspan, (const xmlChar *)"id");
			char *name = (char *)xmlGetProp(myspan, (const xmlChar *)"name");
			char *configname = (char *)xmlGetProp(myspan, (const xmlChar *)"cfgprofile");
			ftdm_span_t *span = NULL;
			uint32_t span_id = 0;
			unsigned paramindex = 0;

			if (!name && !id) {
				fprintf(stderr, "sangoma isdn tap span missing required attribute 'id' or 'name', skipping ...\n");
				continue;
			}

			if (name) {
				zstatus = ftdm_span_find_by_name(name, &span);
			} else {
				if (ftdm_is_number(id)) {
					span_id = atoi(id);
					zstatus = ftdm_span_find(span_id, &span);
				}

				if (zstatus != FTDM_SUCCESS) {
					zstatus = ftdm_span_find_by_name(id, &span);
				}
			}

			if (zstatus != FTDM_SUCCESS) {
				fprintf(stderr, "Error finding FreeTDM span id:%s name:%s\n", id, name);
				continue;
			}

			if (!span_id) {
				span_id = ftdm_span_get_id(span);
			}

			memset(spanparameters, 0, sizeof(spanparameters));
			paramindex = 0;


			if (configname) {
				paramindex = add_isdn_tap_profile_parameters(cfg, configname, spanparameters, ftdm_array_len(spanparameters));
				if (paramindex) {
					fprintf(stdout, "Added %d parameters from profile %s for span %d\n", paramindex, configname, span_id);
				}
			}

			for (param = xml_child(myspan, "param"); param; param = param->next) {
				char *var = (char *) xmlGetProp(param, (xmlChar *)"name");
				char *val = (char *) xmlGetProp(param, (xmlChar *)"value");

				if (var == NULL) {
					continue;
				}
				printf ("PUSHKAR: var = %s and val = %s\n", var, val);
				if (ftdm_array_len(spanparameters) - 1 == paramindex) {
					fprintf(stderr, "Too many parameters for ISDN tap span, ignoring any parameter after %s\n", var);
					break;
				}

				spanparameters[paramindex].var = var;
				spanparameters[paramindex].val = val;
				paramindex++;
			}

			if (ftdm_configure_span_signaling(span, "sangoma_isdn_tap", on_isdn_tap_signal, spanparameters) != FTDM_SUCCESS) {
				fprintf(stderr, "Error configuring ss7 FreeTDM span %d\n", span_id);
				break;
			}
			SPANS[span_id].inuse = 1;
			SPANS[span_id].ftdm_span = span;
			ftdm_span_start(span);

			span_id++;
		}
	}

	xmlFreeDoc(doc);

	xmlCleanupParser();

	return 0;
}

#define INC_ARG(arg_i) \
	arg_i++; \
	if (arg_i >= argc) { \
		fprintf(stderr, "No option value was given for option %s (argi = %d, argc = %d)\n", argv[arg_i-1], arg_i, argc); \
		exit(1); \
	}

/**********************************************************************************
 * Fun   : main()
 *
 * Desc  : Configures all ports and spans as required by user and provide required
 * 	   platform for tapping calls 
 *
 * Ret   : err | FTDM_SUCCESS
**********************************************************************************/
int main(int argc, char *argv[])
{
	/* local variables initialization */
	char *conf_location = NULL;
	int arg_i = 0;

	memset(SPANS, 0, FTDM_MAX_SPANS_INTERFACE*(sizeof(span_data_t)));
#ifndef __WINDOWS__
	add_signal_handlers(SIGINT);
	add_signal_handlers(SIGTERM);
	add_signal_handlers(SIGKILL);
	add_signal_handlers(SIGUSR1);
	add_signal_handlers(SIGUSR2);
#endif

	for (arg_i = 1; arg_i < argc; arg_i++) {
		if (!strcasecmp(argv[arg_i], "-conf")) {
			INC_ARG(arg_i);
			conf_location = argv[arg_i];
		}   
	}   

	if (!conf_location) {
		fprintf(stderr, "Specify the -conf option with the XML configuration path\n");
		exit(1);
	}

	/* Set basic debuggin parameters */
	ftdm_global_set_logger(sample_tap_app_logger);

	if (FTDM_FAIL == app_open_log_file()) {
		fprintf(stderr, "Failed to open log file\n");
		return -1;
	}

	if (ftdm_global_init() != FTDM_SUCCESS) {
		fprintf(stderr, "Error loading FreeTDM\n");
		exit(1);
	}

	if (ftdm_global_configuration() != FTDM_SUCCESS) {
		fprintf(stderr, "Error loading FreeTDM base configuration\n");
		exit(1);
	}

	/* load application specific configuration */
	if (load_config(conf_location)) {
		fprintf(stderr, "Failed to load ISDN Tap configurations\n");
		ftdm_global_destroy();
		return -1; 
	}

	/* At this point all sangoma hw ports have been configured and started.
	   And FreeTDM Spans have been created and started on each hw port */
	ftdm_sleep(1000);

	/* Enable Stats on all FTDM Span and Channels */
	sample_app_exec_io_stats("enable");

	/* Setting D channel wanpipe queue size */
	sample_app_set_rx_queuesize(900); 

	/* mark that initialization done */
	g_init=1;

	/* User drive main menu. 
	 * Test features for the Sangoma TDM SDK 
	 */
	ftdm_log(FTDM_LOG_INFO, "Sangoma PRI/BRI TAPPING MODULE IN ACTION\n\n\n");
	while (tapping_main_menu()) {
		continue;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : exit_sample_app()
 *
 * Desc  : Stops all FTDM as well as application thread and deallocate the memory
 * 	   stop all the configured ports and exit from application.
 *
 * Ret   : NULL
**********************************************************************************/
int exit_sample_app(int max_wait)
{

	/* Stop all threads. 
	 * TODO: use join instead */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		ftdm_log(FTDM_LOG_INFO, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	if (g_init) {
		/* Stop FreeTDM Freamwork and all Spans */
		ftdm_global_destroy();
	}

	ftdm_sleep(1000);

	/* close logging file */
	if (log_file_ptr) {
		fclose(log_file_ptr);
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : open_audio_file()
 *
 * Desc  : Open the the file to which raw data needs to  be written in order to tap
 * 	   the call.
 *
 * Ret   : -1 | fd
**********************************************************************************/
int open_audio_file(uint32_t chan_id, uint32_t span_id, void *obj)
{
	int fd = 0;
	char *path = (char *)obj;

	if (!path) {
		ftdm_log(FTDM_LOG_ERROR, "Cannot Open path as it is empty\n");
		return -1;
	}

	if ((fd = open(path, O_WRONLY | O_CREAT | O_TRUNC
					| FTDM_O_BINARY, S_IRUSR | S_IWUSR)) < 0 ) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to open file [%s], Error[%s]\n", path, strerror(errno));
		return -1;
	}
	ftdm_log(FTDM_LOG_DEBUG, "Start Tapping Call on channel %d:%d to [%s] having fd[%d]\n", span_id, chan_id, path, fd);
	return fd;
}

/* Applications Logging module ***************************************************/
/**********************************************************************************
 * Fun   : app_open_log_file()
 *
 * Desc  : Get log file path and name from user else use default and open same in
 * 	   append mode so that logger can write in to it
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
ftdm_status_t app_open_log_file()
{
	char file_path[MAX_FILE_PATH_SIZE] = { 0 };

	/* get log file path and name and open the same in order to direct logs in to it*/
	fprintf(stdout, "Please Enter the file name with path in which SDK logs needs to be written: ");
	if (NULL != fgets(file_path, MAX_FILE_PATH_SIZE, stdin)) {
		file_path[strlen(file_path) - 1] = '\0';
		if (strlen(file_path)) {
			fprintf(stdout, "SDK log file directory: %s\n", file_path);
		} else {
			strcpy(file_path, "logs/sdk.log");
			fprintf(stdout, "Creating default i.e. %s file in current directory\n", file_path);
		}
	} else {
		strcpy(file_path, "logs/sdk.log");
		fprintf(stdout, "Creating default i.e. %s file in current directory\n", file_path);
	}

	fprintf(stdout, "Please press ENTER to continue...\n");
	getchar();

	if (file_path != NULL) {
		/* open file to write Debug prints i.e. logging in to a file */
		if ((log_file_ptr = fopen(file_path, "a")) == NULL) {
			fprintf(stderr, "Failed to open %s file\n", file_path);
			return FTDM_FAIL;
		}
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : app_logger()
 *
 * Desc  : This is application logger and mainly print logs on screen & file
 *
 * Ret   : void
**********************************************************************************/
void sample_tap_app_logger(const char *file, const char *func, int line, int level, const char *fmt, ...) 
{
	char data[1024] = { 0 }; 
	char curr_time[80] = { 0 }; 
	va_list ap;
	const char *filep = (file ? sangoma_freetdm_cut_path(file) : ""); 

	if (level < 0 || level > 7) { 
		level = 7; 
	}    

	va_start(ap, fmt);

	vsnprintf(data, sizeof(data), fmt, ap); 

	/* get current date and time */
	sangoma_get_timestamp_str(curr_time);

	/* Avoid dumping everything to shell ..let that dump to a log file*/

#ifndef SDK_DEBUGGING
	/* Dump only critical errors logging */
	if ((!strcasecmp("EMERG", SANGOMA_FTDM_LEVEL_NAMES[level])) ||
			(!strcasecmp("CRIT", SANGOMA_FTDM_LEVEL_NAMES[level])) ||
			(!strcasecmp("ERROR", SANGOMA_FTDM_LEVEL_NAMES[level])) ||
			(!strcasecmp("WARNING", SANGOMA_FTDM_LEVEL_NAMES[level]))||
			(!strcasecmp("NOTICE", SANGOMA_FTDM_LEVEL_NAMES[level]))) {
		fprintf(stderr, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data);
	}
#else
	fprintf(stdout, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data);
#endif

	/* dump logs in to file */
	if (log_file_ptr) {
		fprintf(log_file_ptr, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data);
	}

	va_end(ap);
}
/*********************************************************************************/

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4:
 */
